/*
 * DO NOT EDIT.  THIS FILE IS GENERATED FROM /builds/moz2_slave/mozilla-1.9.1-linux-xulrunner/build/gfx/idl/gfxtypes.idl
 */

#ifndef __gen_gfxtypes_h__
#define __gen_gfxtypes_h__


#ifndef __gen_nsrootidl_h__
#include "nsrootidl.h"
#endif

/* For IDL files that don't want to include root IDL files. */
#ifndef NS_NO_VTABLE
#define NS_NO_VTABLE
#endif
/**
 * A color is a 32 bit unsigned integer with
 * four components: R, G, B and A.
 *
 * @var typedef PRUint32 gfx_color
 */
typedef PRUint32 gfx_color;

/**
 * typedef that should be used for bit depths
 * @var typedef unsigned short gfx_depth
 */
typedef PRUint16 gfx_depth;

/**
 * typedef that should be used for image formats
 * @var typedef long gfx_format
 * @see gfxIFormats
 */
typedef PRInt32 gfx_format;


#endif /* __gen_gfxtypes_h__ */
